@@ -1,10 +1,10 @@
 #!/usr/lib/pptp-php-gtk/php/bin/php -q
 <?php
 #
-#   $Id: pptp-php-gtk.php,v 1.18 2003/12/16 22:28:50 quozl Exp $
+#   $Id: pptp-php-gtk.php,v 1.20 2004/06/05 12:23:52 quozl Exp $
 #
 #   pptp-php-gtk.php, PPTP configuration and management GUI
-#   Copyright (C) 2002, 2003  James Cameron (quozl@us.netrek.org)
+#   Copyright (C) 2002-2004  James Cameron (quozl@us.netrek.org)
 #
 #   This program is free software; you can redistribute it and/or modify
 #   it under the terms of the GNU General Public License as published by
@@ -24,7 +24,7 @@
 # our program name
 $me = 'pptp-php-gtk';
 $id = "
-# $Id: pptp-php-gtk.php,v 1.18 2003/12/16 22:28:50 quozl Exp $
+# $Id: pptp-php-gtk.php,v 1.20 2004/06/05 12:23:52 quozl Exp $
 ";
 
 # method to use to operate pppd pipe (experimental)
@@ -164,14 +164,12 @@ function open_window($name) {
 # utility functions, hide a window
 function hide_window($widget) {
   $window = $widget->get_toplevel();
-  $name = $window->get_name();
   $window->hide();
 }
 
 # utility functions, kill a window
 function kill_window($widget) {
   $window = $widget->get_toplevel();
-  $name = $window->get_name();
   $window->destroy();
 }
 
@@ -336,7 +334,7 @@ function files_save_peers($tunnel) {
     }
 
     $x = '';
-    $x .= "# tunnel ".$tunnel['name'].", written by $me \$Revision: 1.18 $\n\n";
+    $x .= "# tunnel ".$tunnel['name'].", written by $me \$Revision: 1.20 $\n\n";
     $x .= "# name of tunnel, used to select lines in secrets files\n";
     $x .= "remotename ".munge($tunnel['name'])."\n\n";
     $x .= "# name of tunnel, used to name /var/run pid file\n";
@@ -560,7 +558,8 @@ function on_main_setup_activate($widget) {
 	$clist->set_column_auto_resize(3, TRUE);
 
 	# enable table sort by user, click a column heading to sort
-	# todo: disabled, sort changes row numbers, invalidating $rows[] array
+	# todo: reimplement.  it is currently disabled because row
+	# numbers are changed by sort, invalidating $rows[] array
 	# $clist->connect('click_column', 'clist_click_column');
 	# $clist->set_reorderable(TRUE);
 
@@ -594,7 +593,7 @@ function setup_initialise() {
 	      # text entry fields setup notebook
 	      'texts' => array('name', 'server', 'domain', 'username',
 			       'password', 'pppd-options', 'pptp-options',
-			       'resolv'),
+			       'resolv', 'dns-options'),
 	      # radio buttons on setup notebook
 	      'radios' => array('routing' => array('routing_interface_only',
 						   'routing_client_to_lan', 
@@ -750,7 +749,12 @@ function resolv_start($context, $interface) {
     $fn = '';
 
     if ($tunnel['usepeerdns']) {
-	$fn = '/etc/ppp/resolv.conf';
+	$fn = '/var/run/ppp/resolv.conf';
+	if (!@is_readable($fn)) $fn = '/etc/ppp/resolv.conf';
+	if (!@is_readable($fn)) {
+	    $fn = '';
+	    scribe($context, $me.": usepeerdns was set, but /{var/run,etc}/ppp/resolv.conf was not readable\n");
+	}
     } else {
 	if ($tunnel['resolv'] != '') {
 	    $fn = '/etc/resolv.conf.'.munge($tunnel['name']);
@@ -764,6 +768,16 @@ function resolv_start($context, $interface) {
     }
 
     if ($fn != '') {
+
+	if ($tunnel['dns-options'] != '') {
+	    $fp = fopen($fn, 'a');
+	    $options = split(';', $tunnel['dns-options']);
+	    foreach ($options as $option) {
+		fwrite($fp, $option."\n");
+	    }
+	    fclose($fp);
+	}
+
 	if (!@rename($on, $sn)) {
 	    scribe($context, $me.": failed to save $on as $sn\n");
 	    scribe($context, $me.": DNS changes were not done\n");
@@ -798,6 +812,15 @@ function routing_start($context, $interface) {
     $name = $context['name'];
     $tunnel = $tunnels[$name];
 
+    # add a route to the server via the interface it used to require
+    $server_ip = $context['ip'];
+    if ($server_ip != '') {
+	$command = 'ip route add '.$context['route'];
+	$array = command($context, $command);
+	if ($array['status'] != 0) $ok = 0;
+	$undo[] = ereg_replace('route add', 'route del', $command);
+    }
+
     # if interface only routing is requested, do nothing
     # http://pptpclient.sourceforge.net/routing.phtml#client-to-server
     if ($tunnel['routing'] == 'routing_interface_only') return;
@@ -835,7 +858,7 @@ function routing_start($context, $interface) {
 	    # obtain current default route for later restoration
 	    list(,$default) = split('  *',`route -n | grep " UG "`);
 
-	    $command = 'route del default '.$public;
+	    $command = 'route del default dev '.$public;
 	    $array = command($context, $command);
             # if there was a default route to the public interface
 	    if ($array['status'] == 0) {
@@ -990,7 +1013,6 @@ function setup_list_insert($tunnel) {
     $row = $clist->append(setup_list_encode($tunnel));
     $key = $tunnel['name'];
     $rows[$key] = $row;
-    echo "setup_list_insert: set \$rows['$key'] to $row\n";
 }
 
 function setup_list_set_state($name, $state) {
@@ -1014,7 +1036,6 @@ function setup_list_set_state($name, $state) {
     $style->base[GTK_STATE_NORMAL] = $normal;
     $style->bg[GTK_STATE_SELECTED] = $selected;
     $key = $name;
-    echo "setup_list_set_state: read \$rows['$key'] as $rows[$key]\n";
     $clist->set_text($rows[$key], 3, $state);
     $clist->set_row_style($rows[$key], $style);
 }
@@ -1027,10 +1048,8 @@ function setup_list_replace($tunnel, $row) {
 
     $key = $clist->get_text($row, 0);
     unset($rows[$key]);
-    echo "setup_list_replace: unset \$rows['$key']\n";
     $key = $name;
     $rows[$key] = $row;
-    echo "setup_list_replace: set \$rows['$key'] to $row\n";
 
     $array = setup_list_encode($tunnel);
     $column = 0;
@@ -1110,7 +1129,7 @@ function on_setup_press_row($widget, $event) {
 	    $gx['setup-popup'] = open_window('pptp-php-gtk-setup-popup');
 	    $menu = $gx['setup-popup']->get_widget('pptp-php-gtk-setup-popup');
 	    $menu->popup(NULL, NULL, NULL, $event->button, $event->time);
-            # todo: ??? need to destroy this on popdown
+            # todo: need to destroy this on popdown
         }
     }
 
@@ -1246,6 +1265,8 @@ function scribe($context, $line) {
     if (ereg('^primary   DNS address', $line)) $colour = $colours['ok'];
     if (ereg('^secondary DNS address', $line)) $colour = $colours['ok'];
     if (ereg('failed', $line)) $colour = $colours['uh'];
+    if (ereg('^anon warn', $line)) $colour = $colours['uh'];
+    if (ereg('^anon fatal', $line)) $colour = $colours['uh'];
     if (ereg('^'.$me, $line)) $colour = $colours['me'];
     
     $context['text']->insert(NULL, $colour, NULL, $line);
@@ -1278,6 +1299,10 @@ function reader($ignore, $ignore, $context) {
     	if (!feof($context['pipe'])) return TRUE;
     }
 
+    if (tunnel_name_to_debug($context['name'])) {
+	scribe($context, "# route -n (after pppd exit)\n" . `route -n`);
+    }
+
     if ($pipe_method == 'fifo') {
 	$status = fclose($context['pipe']);
 	unlink($context['fifo']);
@@ -1352,6 +1377,10 @@ function reader($ignore, $ignore, $context) {
     }
     gtk::input_remove($window->get_data('input'));
 
+    if (tunnel_name_to_debug($context['name'])) {
+	scribe($context, "# route -n (after completion)\n" . `route -n`);
+    }
+
     return FALSE;
 }
 
@@ -1420,6 +1449,19 @@ function tunnel_window_create($name) {
     $window->set_data('ping', $ping);
     $ping->set_data('name', $name);
 
+    $copy = $gx[$name]->get_widget('tunnel_copy');
+    $window->set_data('tunnel_copy', $copy);
+    $copy->set_data('name', $name);
+    # todo: this feature doesn't work properly yet, see on_tunnel_copy_activate
+    $copy->hide();
+
+    $save_as = $gx[$name]->get_widget('tunnel_save_as');
+    $window->set_data('tunnel_save_as', $save_as);
+    $save_as->set_data('name', $name);
+
+    $close = $gx[$name]->get_widget('tunnel_close');
+    $close->set_data('name', $name);
+
     return $window;
 }
 
@@ -1515,8 +1557,26 @@ kill() system call returned an error for PID '$pid'");
     setup_unselect_all();
 }
 
+# erase security critical portions when dumping tunnel configuration
+function tunnel_name_to_hidden_dump($name) {
+    global $tunnels;
+
+    $tunnel = $tunnels[$name];
+    $hidden = '(hidden by pptp-php-gtk)';
+    if ($tunnel['password'] != '') $tunnel['password'] = $hidden;
+    if ($tunnel['domain']   != '') $tunnel['domain']   = $hidden;
+    return print_r($tunnel, true);
+}
+
+function tunnel_name_to_debug($name) {
+    global $tunnels;
+
+    $tunnel = $tunnels[$name];
+    return $tunnel['debug'];
+}
+
 function start($name) {
-    global $selected, $tunnels, $gx, $pipe_method, $marker;
+    global $selected, $tunnels, $gx, $pipe_method, $marker, $me;
 
     # check for an existing tunnel
     $pidfile = pid_name($name);
@@ -1552,6 +1612,36 @@ or $pidfile has been left.");
 	$text = $gx[$name]->get_widget('text');
 	$context = array('name' => $name, 'tree' => $gx[$name], 'window' => $window, 'text' => $text);
 
+	# obtain the route to the server before pppd starts, so that
+	# it can be used to add a host route after pppd has connected
+	$tunnel = $tunnels[$name];
+	$server_name = $tunnel['server'];
+	$server_ip = @gethostbyname($server_name);
+	if ($server_ip != '') {
+	    $context['ip'] = $server_ip;
+	    $context['route'] = trim(`ip route get $server_ip | head -1`);
+	}
+	
+	# todo: gethostbyname() may return a different value if the
+	# tunnel's host name is a round-robin DNS entry, which may
+	# invalidate this route.
+
+	# in debug mode, dump kernel version, module names, module info
+	if (tunnel_name_to_debug($context['name'])) {
+	    scribe($context, "$me: debug information dump begins\n");
+	    scribe($context, "WARNING: security sensitive information follows\n");
+	    scribe($context, id() . "\n");
+	    scribe($context, "# pppd --version\n" . `pppd --version 2>&1`);
+	    scribe($context, "# uname -a\n" . `uname -a`);
+	    scribe($context, "# grep mppe /proc/modules\n" . `grep mppe /proc/modules`);
+	    scribe($context, "# modinfo ppp_mppe\n" . `modinfo ppp_mppe`);
+	    scribe($context, tunnel_name_to_hidden_dump($context['name']));
+	    scribe($context, "# route -n (before pppd)\n" . `route -n`);
+	    # todo: additional configuration details 
+	    # pptp-linux version (not yet supported by pptp-linux!)
+	    scribe($context, "$me: debug information dump ends, starting pppd\n");
+	}
+
 	# construct the pppd command, with log output to us, detach on IP up
 	$command = '/usr/sbin/pppd call '.munge($name).' logfd 1 updetach 2>&1;';
 	# save the pppd exit code into a shell variable
@@ -1578,6 +1668,8 @@ or $pidfile has been left.");
 	    # socket_set_blocking($context['pipe'], FALSE);
 	}
 
+# todo: review the observations below for newer versions of php/php-gtk
+
 # todo: neither method works properly, still get situations where the
 # pipe should be ready to be read but the input function is not been
 # called.  we can't try to read it in a timeout, because there are no
@@ -1943,6 +2035,13 @@ function on_tunnel_close_activate($widget) {
     hide_window($widget);
 }
 
+function on_tunnel_close_clicked($widget) {
+    global $gx;
+    $name = $widget->get_data('name');
+    $window = $gx[$name]->get_widget('pptp-php-gtk-tunnel');
+    $window->hide();
+}
+
 function on_tunnel_stop_activate($widget) {
     $name = $widget->get_data('name');
     stop($name);
@@ -1963,6 +2062,76 @@ function on_tunnel_delete_event($widget) {
     return TRUE;
 }
 
+function diagnostic_dump($name) {
+    global $gx;
+
+    $text = $gx[$name]->get_widget('text');
+    return $text->get_chars(0, -1);
+}
+
+function on_tunnel_copy_activate($widget) {
+    $name = $widget->get_data('name');
+
+    $entry = &new GtkEntry();
+    $entry->hide();
+    $entry->set_text(diagnostic_dump($name));
+    $entry->select_region(0, -1);
+    $entry->copy_clipboard();
+    # todo: the copy buffer appears to be limited to 2048 bytes!
+    # perhaps the widget needs to be instantiated for the rest of the
+    # paste to work
+
+    # todo: release this poor widget
+}
+
+function on_tunnel_save_as_activate($widget) {
+    $gx = open_window('pptp-php-gtk-save-as');
+    $fs = $gx->get_widget('pptp-php-gtk-save-as');
+
+    $ok = $fs->ok_button;
+    $ok->connect('clicked', 'on_tunnel_save_as_ok', $fs);
+    $name = $widget->get_data('name');
+    $ok->set_data('name', $name);
+
+    $cancel = $fs->cancel_button;
+    $cancel->connect('clicked', 'on_tunnel_save_as_cancel', $fs);
+}
+
+function on_tunnel_save_as_ok($button, $fs) {
+    $fn = $fs->get_filename();
+    $name = $button->get_data('name');
+    $fp = fopen($fn, 'w');
+    if (!$fp) {
+	message('tunnel save as ok', "Cannot write to file $fn");
+	return;
+    }
+    if (!fwrite($fp, diagnostic_dump($name))) {
+	message('tunnel save as write', "Failed to write to $fn after opening it for write
+information has not been saved.");
+	fclose($fp);
+	return;
+    }
+    if (!fclose($fp)) {
+	message('tunnel save as write', "Failed to close $fn after writing
+information may not have been saved.");
+	fclose($fp);
+	return;
+    }
+    $window = $button->get_toplevel();
+    kill_window($window);
+    message("security warning", "File $fn saved.\n\nThe file may contain information that would allow an attacker to use your tunnel service rights.
+
+Before posting this publically, consistently change:
+- your username,
+- your authentication domain name,
+- your server name or IP addresses, and
+- your IP addresses (shown in routing table dumps).");
+}
+
+function on_tunnel_save_as_cancel($button, $fs) {
+    $window = $button->get_toplevel();
+    kill_window($window);
+}
 
 # window pptp-php-gtk-message 
 function message($title, $text) {
